// ------------------------------------------------------------
// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.
// ------------------------------------------------------------

package redis

import (
	"crypto/tls"
	"errors"
	"fmt"
	"strconv"
	"time"

	"github.com/dapr/components-contrib/pubsub"
	"github.com/dapr/dapr/pkg/logger"
	"github.com/go-redis/redis/v7"
)

const (
	host       = "redisHost"
	password   = "redisPassword"
	consumerID = "consumerID"
	enableTLS  = "enableTLS"
)

type redisStreams struct {
	metadata metadata
	client   *redis.Client

	logger logger.Logger
}

// NewRedisStreams returns a new redis streams pub-sub implementation
func NewRedisStreams(logger logger.Logger) pubsub.PubSub {
	return &redisStreams{logger: logger}
}

func parseRedisMetadata(meta pubsub.Metadata) (metadata, error) {
	m := metadata{}
	if val, ok := meta.Properties[host]; ok && val != "" {
		m.host = val
	} else {
		return m, errors.New("redis streams error: missing host address")
	}

	if val, ok := meta.Properties[password]; ok && val != "" {
		m.password = val
	}

	if val, ok := meta.Properties[enableTLS]; ok && val != "" {
		tls, err := strconv.ParseBool(val)
		if err != nil {
			return m, fmt.Errorf("redis streams error: can't parse enableTLS field: %s", err)
		}
		m.enableTLS = tls
	}

	if val, ok := meta.Properties[consumerID]; ok && val != "" {
		m.consumerID = val
	} else {
		return m, errors.New("redis streams error: missing consumerID")
	}

	return m, nil
}

func (r *redisStreams) Init(metadata pubsub.Metadata) error {
	m, err := parseRedisMetadata(metadata)
	if err != nil {
		return err
	}
	r.metadata = m

	options := &redis.Options{
		Addr:            m.host,
		Password:        m.password,
		DB:              0,
		MaxRetries:      3,
		MaxRetryBackoff: time.Second * 2,
	}

	/* #nosec */
	if r.metadata.enableTLS {
		options.TLSConfig = &tls.Config{
			InsecureSkipVerify: r.metadata.enableTLS,
		}
	}

	client := redis.NewClient(options)

	_, err = client.Ping().Result()
	if err != nil {
		return fmt.Errorf("redis streams: error connecting to redis at %s: %s", m.host, err)
	}

	r.client = client

	return nil
}

func (r *redisStreams) Publish(req *pubsub.PublishRequest) error {
	_, err := r.client.XAdd(&redis.XAddArgs{
		Stream: req.Topic,
		Values: map[string]interface{}{"data": req.Data},
	}).Result()
	if err != nil {
		return fmt.Errorf("redis streams: error from publish: %s", err)
	}

	return nil
}

func (r *redisStreams) Subscribe(req pubsub.SubscribeRequest, handler func(msg *pubsub.NewMessage) error) error {
	err := r.client.XGroupCreateMkStream(req.Topic, r.metadata.consumerID, "0").Err()
	if err != nil {
		r.logger.Warnf("redis streams: %s", err)
	}
	go r.beginReadingFromStream(req.Topic, r.metadata.consumerID, handler)

	return nil
}

func (r *redisStreams) readFromStream(stream, consumerID, start string) ([]redis.XStream, error) {
	res, err := r.client.XReadGroup(&redis.XReadGroupArgs{
		Group:    consumerID,
		Consumer: consumerID,
		Streams:  []string{stream, start},
		Block:    0,
	}).Result()
	if err != nil {
		return nil, err
	}

	return res, nil
}

func (r *redisStreams) processStreams(consumerID string, streams []redis.XStream, handler func(msg *pubsub.NewMessage) error) {
	for _, s := range streams {
		for _, m := range s.Messages {
			go func(stream string, message redis.XMessage) {
				msg := pubsub.NewMessage{
					Topic: stream,
				}
				data, exists := message.Values["data"]
				if exists && data != nil {
					msg.Data = []byte(data.(string))
				}

				err := handler(&msg)
				if err == nil {
					r.client.XAck(stream, consumerID, message.ID).Result()
				}
			}(s.Stream, m)
		}
	}
}

func (r *redisStreams) beginReadingFromStream(stream, consumerID string, handler func(msg *pubsub.NewMessage) error) {
	// first read pending items in case of recovering from crash
	start := "0"

	for {
		streams, err := r.readFromStream(stream, consumerID, start)
		if err != nil {
			r.logger.Errorf("redis streams: error reading from stream %s: %s", stream, err)

			return
		}
		r.processStreams(consumerID, streams, handler)

		// continue with new non received items
		start = ">"
	}
}

func (r *redisStreams) Close() error {
	return r.client.Close()
}

func (r *redisStreams) Features() []pubsub.Feature {
	return nil
}
